#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable

from pywaclient.endpoints import CrudEndpoint

class NoteCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'note')



class NoteSectionCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'notesection')
        self.note = NoteCrudEndpoint(client)
        self.note_sections_path = f'{self.path}/notes'

    def notes(self, note_section_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        Retrieve an iterable of all the note references within one section.

        :param note_section_id: The id of the parent note section.
        :param complete:        Ignore the limit and offsets and return all the entities.
        :param limit:           The number of returned references between 1 and 50.
        :param offset:          The offset in the returned list.
        :return:
        """
        if complete:
            return self._scroll_collection(self.note_sections_path, {'id': note_section_id}, 'entities')
        assert offset >= 0
        assert 1 <= limit <= 50
        return self._post_request(self.note_sections_path, {'id': note_section_id}, {'limit': limit, 'offset': offset})['entities']


class NotebookCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'notebook')
        self.note_section = NoteSectionCrudEndpoint(client)
        self.path_notebook_sections = f'{self.path}/notesections'

    def note_sections(self, notebook_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        Retrieve an iterable of all the note section references from one notebook.

        :param notebook_id: Identifier of the parent notebook.
        :param complete:    Ignore the limit and offsets and return all the entities.
        :param limit:       The number of returned references between 1 and 50.
        :param offset:      The offset in the returned list.
        :return:
        """
        if complete:
            return self._scroll_collection(self.path_notebook_sections, {'id': notebook_id}, 'entities')
        assert offset >= 0
        assert 1 <= limit <= 50
        return self._post_request(self.path_notebook_sections, {'id': notebook_id}, {'limit': limit, 'offset': offset})['entities']


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from unittest.mock import Mock

import pytest
from hypothesis import given, strategies as st, settings, example

from pywaclient.endpoints import _parse_response
from pywaclient.exceptions import AccessForbidden, ResourceNotFound, InternalServerException, \
    UnexpectedStatusException, ConnectionException, UnprocessableDataProvided, FailedRequest

key_values = [
    'success',
    'id',
    'title',
    'slug',
    'excerpt',
    'content',
    'templateType',
]

def is_status_ok(status_code):
    return 199 < status_code < 400

def add_error(x: dict):
    x['error'] = st.text()
    return x
@given(
    path=st.text(),
    status_code=st.integers(min_value=200, max_value=599),
    content=st.dictionaries(
        keys=st.sampled_from(key_values),
        values=st.text()).map(lambda x: st.fixed_dictionaries({'success': st.booleans()}).map(lambda fixed_dict: {**fixed_dict, **x}))
)
@settings(max_examples=1000)
@example(path='', status_code=200, content={'success': True})
def test_parse_response(path, status_code, content):
    response = Mock(status_code=status_code, ok=is_status_ok(status_code), json=Mock(return_value=content))
    params = {'param1': 'value1', 'param2': 'value2'}

    if 'success' not in content:
        with pytest.raises(UnexpectedStatusException):
            _parse_response(path, response, params, content)
    if is_status_ok(status_code) and content.get('success', False):
        assert _parse_response(path, response, params, content) == content
    elif is_status_ok(status_code) and not content.get('success', False):
        with pytest.raises(FailedRequest):
            _parse_response(path, response, params, content)
    elif status_code == 401 or status_code == 403:
        with pytest.raises(AccessForbidden):
            _parse_response(path, response, params, content)
    elif status_code == 404:
        with pytest.raises(ResourceNotFound):
            _parse_response(path, response, params, content)
    elif status_code == 422:
        with pytest.raises(UnprocessableDataProvided):
            _parse_response(path, response, params, content)
    elif status_code == 500:
        with pytest.raises(InternalServerException):
            _parse_response(path, response, params, content)
    else:
        with pytest.raises(UnexpectedStatusException):
            _parse_response(path, response, params, content)



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from init_client import client

if __name__ == '__main__':
    auth_user = client.user.identity()
    full_user = client.user.get(auth_user['id'], 2)

    for w in client.user.worlds(full_user['id']):
        world = client.world.get(w['id'], 1)
        print(world)

    for t in client.user.block_templates(full_user['id']):
        print(t)

    for n in client.user.notebooks(auth_user['id']):
        print(n)







#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    for history in client.world.histories(world_id):
        print(history['title'])
        client.history.delete(history['id'])

    test_timeline_1 = client.timeline.put(
        {
            'title': 'Test Timeline Creation',
            'world': {
                'id': world_id
            }
        }
    )

    test_histories_1 = client.history.put(
        {
            'title': 'Test Histories Creation',
            "year":"5545",
            'world': {
                'id': world_id
            }
        }
    )
    test_histories_2 = client.history.put(
        {
            'title': 'Test Histories Creation 2',
            'year': '2992',
            'world': {
                'id': world_id
            },
            # 'timelines': [
            #     {
            #         'id': test_timeline_1['id']
            #     }
            #]
        }
    )
    response_patch_histories_2 = client.history.patch(
        test_histories_2['id'],
        {
            'hour': '2001ad'
        }
    )

    full_test_histories_2 = client.history.get(
        test_histories_2['id'],
        2
    )

    assert full_test_histories_2['hour'] == 2001

    client.history.delete(test_histories_1['id'])
    client.history.delete(test_histories_2['id'])
    client.history.delete(test_timeline_1['id'])

    for history in client.world.histories(world_id):
        client.history.delete(history['id'])



from typing import Dict, Any


class WorldAnvilClientException(Exception):
    """Base exception to the library."""


class WorldAnvilServerException(WorldAnvilClientException):
    """Exceptions returned by the server for requests made."""

    def __init__(self, status: int, message: str, summary: str = "", path: str = None, params: Dict[str, Any] = None, content: Dict[str, Any] = None):
        self.status = status
        self.message = message
        self.summary = summary
        self.path = path
        self.content = content
        self.params = params


class ConnectionException(WorldAnvilServerException):
    """Was unable to connect to World Anvil for some reason."""

    def __init__(self, message: str):
        super().__init__(503, message)


class UnexpectedStatusException(WorldAnvilServerException):
    """An unexpected status exception occurred."""

    def __init__(self, status: int, message: str, path: str, params: Dict[str, Any], content: Dict[str, Any]):
        super().__init__(status, message, "", path, params, content)


class InternalServerException(WorldAnvilServerException):
    """Internal Server Error in World Anvil Response."""

    def __init__(self, status: int, path: str, params: Dict[str, Any], content: Dict[str, Any]):
        super().__init__(status, 'World Anvil was unable to process this request.', "", path, params, content)


class UnauthorizedRequest(WorldAnvilServerException):
    """The user is not authorized to access the requested resource. This can happen if no authentication token or application key was provided or the provided token or key are not valid."""

    def __init__(self, path: str, params: Dict[str, Any], content: Dict[str, Any]):
        super().__init__(401, "Unauthorized.", "Either no authentication token or application key was provided or the provided token or key are not valid.", path, params, content)

class AccessForbidden(WorldAnvilServerException):
    """The user does not have permissions for the requested resources."""

    def __init__(self, path: str, params: Dict[str, Any], content: Dict[str, Any]):
        super().__init__(403, "Permission denied.", "The authentication token does not allow access to the requested resource.", path, params, content)


class ResourceNotFound(WorldAnvilServerException):
    """The requested resource does not exist or was moved."""

    def __init__(self, path: str, params: Dict[str, Any], content: Dict[str, Any]):
        super().__init__(404, "Requested resource was not found.", "", path, params, content)


class UnprocessableDataProvided(WorldAnvilServerException):
    """The request could not be processed."""

    def __init__(self, path: str, data: Dict[str, Any], params: Dict[str, Any], content: Dict[str, Any]):
        if 'status' in data:
            super().__init__(422, "Unprocessable data provided", data['error']['summary'], path, params, content)
            self.error_tracestack = data['error']['traceStack']
        else:
            super().__init__(422, "Unprocessable data provided", data['error'], path, params, content)
            self.error_tracestack = data['trace']

class FailedRequest(WorldAnvilServerException):
    """Status code indicated success, but request failed."""

    def __init__(self, status: int, path: str, message: str, response: Dict[str, Any], params: Dict[str, Any], content: Dict[str, Any]):
        super().__init__(status, message, path, params, content)
        self.response = response


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Dict, Any, Iterable

from pywaclient.endpoints import CrudEndpoint


class WorldCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'world')
        self.path_articles = f'{self.path}/articles'
        self.path_categories = f'{self.path}/categories'
        self.path_block_folders = f'{self.path}/blockfolders'
        self.path_variable_collections = f'{self.path}/variablecollections'
        self.path_subscriber_groups = f'{self.path}/subscribergroups'
        self.path_secrets = f'{self.path}/secrets'
        self.path_manuscripts = f'{self.path}/manuscripts'
        self.path_maps = f'{self.path}/maps'
        self.path_images = f'{self.path}/images'
        self.path_histories = f'{self.path}/histories'
        self.path_timelines = f'{self.path}/timelines'
        self.path_chronicles = f'{self.path}/chronicles'
        self.path_canvases = f'{self.path}/canvases'
        self.path_blocks = f'{self.path}/blocks'

    def articles(self, world_id: str, category_id: str = None, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        List all articles by a world, filtered with a limit of entities shown at an offset and optionally by a category.

        :param world_id:    The id of the world the articles should be returned from.
        :param category_id: (optional) The id of the category to return the articles from. To get articles without a category set
                            this value to -1. This does not return articles that have a parent article instead of a
                            category.
        :param complete     Ignore limit and offset and return all the articles as an iterable. Will fetch a new batch
                            every 50 articles.
        :param limit:       Determines how many articles are returned. Value between 1 and 50.
        :param offset:      Determines the offset at which articles are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            if category_id is None:
                return self._scroll_collection(self.path_articles, {'id': world_id}, 'entities')
            else:
                return self._scroll_collection(self.path_articles, {'id': world_id}, 'entities', 'category', category_id)
        else:
            request_body = {'limit': limit, 'offset': offset}
            if category_id is not None:
                request_body['category'] = {'id': category_id}
            return self._post_request(self.path_articles,
                                      {'id': world_id}, request_body)['entities']

    def categories(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all categories by a world given, filtered with a limit of entities shown and an offset.

        :param world_id:    The id of the world the folders should be returned from.
        :param complete:    Return the entire list of entities as a list. Only loads 50 categories at a time.
        :param limit:       Determines how many folders are returned. Value between 1 and 50.
        :param offset:      Determines the offset at which folders are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            return self._scroll_collection(self.path_categories, {'id': world_id}, 'entities')
        return self._post_request(self.path_categories,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def statblock_folders(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all statblock folders of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of statblock folder entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_block_folders, {'id': world_id}, 'entities')
        return self._post_request(self.path_block_folders,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def variable_collections(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all variable collections of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of variable collection entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_variable_collections, {'id': world_id}, 'entities')
        return self._post_request(self.path_variable_collections,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def subscriber_groups(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all subscriber groups of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of subscriber group entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_subscriber_groups, {'id': world_id}, 'entities')
        return self._post_request(self.path_subscriber_groups,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def secrets(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all secrets of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of secret entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_secrets, {'id': world_id}, 'entities')
        return self._post_request(self.path_secrets,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def manuscripts(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all manuscripts of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of manuscript entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_manuscripts, {'id': world_id}, 'entities')
        return self._post_request(self.path_manuscripts,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def maps(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all maps of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of map entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_maps, {'id': world_id}, 'entities')
        return self._post_request(self.path_maps,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def images(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all images of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of image entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_images, {'id': world_id}, 'entities')
        return self._post_request(self.path_images,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def histories(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all histories of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of history entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_histories, {'id': world_id}, 'entities')
        return self._post_request(self.path_histories,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def timelines(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all timelines of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of timeline entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_timelines, {'id': world_id}, 'entities')
        return self._post_request(self.path_timelines,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def chronicles(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all chronicles of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of chronicle entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_chronicles, {'id': world_id}, 'entities')
        return self._post_request(self.path_chronicles,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def canvases(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all canvases of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of canvas entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_canvases, {'id': world_id}, 'entities')
        return self._post_request(self.path_canvases,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable

from pywaclient.endpoints import CrudEndpoint


class ManuscriptBeatCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client, 'manuscript_beat')


class ManuscriptPartCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client, 'manuscript_part')
        self._path = 'manuscript_part'
        self.beat = ManuscriptBeatCrudEndpoint(client)
        self.path_beats = f'{self.path}/manuscript_beats'

    def beats(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the beats of a manuscript part.

        :param identifier:  Identifier of the manuscript version.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self.path_beats, {'id': identifier}, 'entities')


class ManuscriptPlotCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client, 'manuscript_plot')


class ManuscriptStatCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client, 'manuscript_stat')


class ManuscriptVersionCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client, 'manuscript_version')
        self.part = ManuscriptPartCrudEndpoint(client)
        self.path_parts = f'{self.path}/manuscript_parts'
        self.plot = ManuscriptPlotCrudEndpoint(client)
        self.path_plots = f'{self.path}/manuscript_plots'
        self.stat = ManuscriptStatCrudEndpoint(client)
        self.path_stats = f'{self.path}/manuscript_stats'

    def parts(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the parts of a manuscript version.

        :param identifier:  Identifier of the manuscript version.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self.path_parts, {'id': identifier}, 'entities')

    def plots(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the plots of a manuscript version.

        :param identifier:  Identifier of the manuscript version.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self.path_plots, {'id': identifier}, 'entities')

    def stats(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the stats of a manuscript version.

        :param identifier:  Identifier of the manuscript version.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self.path_stats, {'id': identifier}, 'entities')


class ManuscriptBookmarkCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'manuscript_bookmark')


class ManuscriptTagCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'manuscript_tag')


class ManuscriptLabelCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'manuscript_label')


class ManuscriptCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'manuscript')
        self.bookmark = ManuscriptBookmarkCrudEndpoint(client)
        self._path_bookmarks = f'{self.path}/manuscript_bookmarks'
        self.tag = ManuscriptTagCrudEndpoint(client)
        self._path_tags = f'{self.path}/manuscript_tags'
        self.label = ManuscriptLabelCrudEndpoint(client)
        self._path_labels = f'{self.path}/manuscript_labels'
        self.version = ManuscriptVersionCrudEndpoint(client)
        self._path_versions = f'{self.path}/manuscript_versions'

    def bookmarks(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the bookmarks on a manuscript.

        :param identifier:  Identifier of the manuscript.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self._path_bookmarks, {'id': identifier}, 'entities')

    def labels(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the labels defined in a manuscript.

        :param identifier:  Identifier of the manuscript.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self._path_labels, {'id': identifier}, 'entities')

    def tags(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the tags added to a manuscript.

        :param identifier:  Identifier of the manuscript.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self._path_tags, {'id': identifier}, 'entities')

    def versions(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Retrieve an iterable of all the versions of a manuscript.

        :param identifier:  Identifier of the manuscript.
        :return:            Iterable dictionary of the entities.
        """
        return self._scroll_collection(self._path_versions, {'id': identifier}, 'entities')



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json

from init_client import client

if __name__ == '__main__':


    test_map = client.map.get('58c3ac4a-6975-483c-b8ad-b81a37178562', 2)
    print(json.dumps(test_map, indent='    ', ensure_ascii=False))

    client.map.patch(test_map['id'],
                     {
                          'isDeleted': True,
                          'displayOnHomepage': False,
                          'views': 10000000,
                      }
                     )


    test_map = client.map.get('58c3ac4a-6975-483c-b8ad-b81a37178562', 2)
    print(json.dumps(test_map, indent='    ', ensure_ascii=False))


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable

from pywaclient.endpoints import CrudEndpoint


class MapMarkerTypeCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'markertype')
        self.path_list = 'markertypes'

    def list(self) -> Iterable[Dict[str, Any]]:
        """Iterate over all map marker types.
        """
        return self._scroll_collection(self.path_list, {}, 'entities')



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    for collection in client.world.variable_collections(world_id):
        print(collection)
        client.variable_collection.get(collection['id'], 2)
        for variable in client.variable_collection.variables(collection['id']):
            client.variable_collection.variable.get(variable['id'], 2)
            print(variable)




    test_variableCollection_1 = client.variable_collection.put(
        {
            "title": "Variables Collection",
            "description": "All variables used in this world.",
            "world": {
                "id": world_id
            },
            "prefix": "test-prefix",
            "state": "public"
        }
    )
    test_variableCollection_2 = client.variable_collection.put(
        {
            "title": "Variables Collection 2",
            "description": "All variables used in this world.",
            "world": {
                "id": world_id
            },
            "prefix": "test-prefix-2",
            "state": "public"
        }
    )
    response_patch_variableCollection_2 = client.variable_collection.patch(
        test_variableCollection_2['id'],
        {
            'description': 'Update this variable collection with a new excerpt.'
        }
    )

    full_test_variableCollection_2 = client.variable_collection.get(
        test_variableCollection_2['id'],
        2
    )

    assert full_test_variableCollection_2['description'] == 'Update this variable collection with a new excerpt.'

    for i in range(50):
        test_variable_1 = client.variable.put(
            {
                "k": str(i),
                "t": "title",
                "v": "value",
                "type": "string",
                "collection": test_variableCollection_1['id'],
                "world": world_id
            }
        )

    for x in client.variable_collection.variables(world_id, test_variableCollection_1['id']):
        print(x)

    client.variableCollection.delete(test_variableCollection_1['id'])
    client.variable_collection.delete(test_variableCollection_2['id'])



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import client, world_id, user_id

if __name__ == '__main__':

    for n in client.user.notebooks(user_id):
        print(n)

    notebook = client.notebook.put(
        {
            'title': 'A New Notebook',
            'world': {
                'id': world_id
            }
        }
    )
    value = client.notebook.get(notebook['id'], 1)
    print(value)

    note_section = client.notebook.note_section.put(
         {
             'title': 'A new section',
             'notebook': {
                 'id': notebook['id']
             }
         }
    )
    print(client.notebook.note_section.get(note_section['id'], 2))

    note = client.notebook.note_section.note.put(
         {
             'title': "A New Note",
             'content': 'Some content',
             'notesection': {
                 'id': note_section['id']
             }
         }
     )
    print(client.notebook.note_section.note.get(note['id'], 2))

    client.notebook.note_section.note.delete(note['id'])
    client.notebook.note_section.delete(note_section['id'])
    client.notebook.delete(notebook['id'])



#    Copyright 2020 - present Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the 'License');
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an 'AS IS' BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import client, world_id

if __name__ == '__main__':
    test_category_1 = client.category.put(
        {
            'title': 'Test Creation Category',
            'world': {
                'id': world_id
            }
        }
    )
    print(test_category_1)
    test_category_2 = client.category.put(
        {
            'title': 'Test Creation Category 2',
            'world': {
                'id': world_id
            }
        }
    )
    print(test_category_2)
    response_patch_article_2 = client.category.patch(
        test_category_2['id'],
        {
            'excerpt': 'This is an excerpt for an article.'
        }
    )
    print(response_patch_article_2)
    full_test_category_2 = client.category.get(
        test_category_2['id'],
        2
    )
    print(full_test_category_2)
    assert full_test_category_2['excerpt'] == 'This is an excerpt for an article.'
    list_of_articles_no_category = client.category.articles(world_id, "-1")
    print(list_of_articles_no_category)

    client.category.delete(test_category_1['id'])
    client.category.delete(test_category_2['id'])



#    Copyright 2020 - present Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the 'License');
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an 'AS IS' BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json

from init_client import world_id, client

if __name__ == '__main__':


    alliance = 'e6df2970-7338-41d6-87af-2eef64ad365a'




    full_test_article = client.article.get(
        alliance,
        2
    )

    print(json.dumps(full_test_article, indent=4, sort_keys=True))





#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import logging
from typing import Dict, Any, Iterable

import requests
from requests import RequestException, Response

from pywaclient.exceptions import AccessForbidden, ResourceNotFound, InternalServerException, \
    UnexpectedStatusException, ConnectionException, UnprocessableDataProvided, FailedRequest, UnauthorizedRequest


def download_binary(url: str, chunk_size: int = 8192) -> Iterable[bytes]:
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        for chunk in r.iter_content(chunk_size=chunk_size):
            yield chunk

def _parse_response(path: str, response: Response, params: Dict[str, Any], content: Dict[str, Any]) -> Dict[str, Any]:
    if response.ok:
        data = response.json()
        if 'success' not in data:
            raise UnexpectedStatusException(response.status_code, 'Response contained no success flag.', path, params, content)
        if data['success']:
            return data
        else:
            raise FailedRequest(response.status_code, path, data['error'], data, params, content)
    elif response.status_code == 401:
        raise UnauthorizedRequest(path, params, content)
    elif response.status_code == 403:
        raise AccessForbidden(path, params, content)
    elif response.status_code == 404:
        raise ResourceNotFound(path, params, content)
    elif response.status_code == 422:
        raise UnprocessableDataProvided(path, response.json(), params, content)
    elif response.status_code == 500:
        raise InternalServerException(500, path, params, content)
    else:
        raise UnexpectedStatusException(response.status_code, response.reason, path, params, content)


class BasicEndpoint:

    def __init__(self, client: 'BoromirApiClient', base_path: str):
        self.client = client
        self.path = base_path

    def _get_request(self, path: str, params: Dict[str, str]) -> Dict[str, Any]:
        try:
            response = requests.get(self.client.base_url + path, params=params, headers=self.client.headers)
            return _parse_response(path, response, params, {})
        except RequestException as err:
            raise ConnectionException(str(err))

    def _put_request(self, path: str, content: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.put(self.client.base_url + path, json=content, headers=self.client.headers_post)
            return _parse_response(path, response, {}, content)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _patch_request(self, path: str, params: Dict[str, str], content: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.patch(self.client.base_url + path, params=params, json=content,
                                      headers=self.client.headers_post)
            return _parse_response(path, response, params, content)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _post_request(self, path: str, params: Dict[str, str], content: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.post(self.client.base_url + path, params=params, json=content,
                                     headers=self.client.headers_post)
            return _parse_response(path, response, params, content)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _delete_request(self, path: str, params: Dict[str, str]):
        try:
            response = requests.delete(self.client.base_url + path,
                                       params=params,
                                       headers=self.client.headers)
            return _parse_response(path, response, params, {})
        except RequestException as err:
            raise ConnectionException(str(err))


    def _scroll_collection(self, path: str, params: Dict[str, str], collection_tag: str, parentName: str = '', parentId: str = '') -> Iterable[Dict[str, Any]]:
        limit = 50
        offset = 0
        if parentName != '':
            content = {'limit': limit, 'offset': offset, parentName: {'id': parentId}}
        else:
            content = {'limit': limit, 'offset': offset}
        collection = self._post_request(path, params, content)
        if collection['success']:
            items = collection[collection_tag]
            while len(items) > 0:
                for item in items:
                    yield item
                offset = offset + limit
                content['offset'] = offset
                collection = self._post_request(path, params, content)
                items = collection[collection_tag]

    def get(self, identifier: str, granularity: int) -> Dict[str, Any]:
        """Retrieve the metadata of the resources. Use the granularity parameter to determine the amount of information
        returned by the api.

        :param identifier:  The id of the resource.
        :param granularity: The granularity of the response. -1: reference only, 0: default, 1: full, 2: extended, 3: special
        :return:            The metadata of the resource.
        """
        return self._get_request(self.path, {'id': identifier, 'granularity': str(granularity)})


class CrudEndpoint(BasicEndpoint):

    def __init__(self, client: 'BoromirApiClient', base_path: str):
        super().__init__(client, base_path)

    def put(self, content: Dict[str, Any]) -> Dict[str, Any]:
        """Create the resource.

        :param content:     The content of the resource. Including required fields.
        :return:            The reference of the created resource.
        """
        return self._put_request(self.path, content)

    def patch(self, identifier: str, content: Dict[str, Any]) -> Dict[str, Any]:
        """Update the resource with the given id. The content should contain the fields that should be overwritten.

        :param identifier:  The id of the resource to be updated.
        :param content:     The fields that should be overwritten.
        :return:            The reference of the updated resource.
        """
        return self._patch_request(self.path, {'id': identifier}, content)

    def delete(self, identifier: str):
        """Delete the resource with the given id.

        :param identifier:  The id of the resource to be deleted.
        """
        self._delete_request(self.path, {'id': identifier})



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from pywaclient.endpoints.articles import ArticleCrudEndpoint
from pywaclient.endpoints.block_templates import BlockTemplateCrudEndpoint
from pywaclient.endpoints.blocks import BlockCrudEndpoint, BlockFolderCrudEndpoint
from pywaclient.endpoints.canvas import CanvasCrudEndpoint
from pywaclient.endpoints.categories import CategoryCrudEndpoint
from pywaclient.endpoints.chronicles import ChronicleCrudEndpoint
from pywaclient.endpoints.histories import HistoryCrudEndpoint
from pywaclient.endpoints.images import ImageCrudEndpoint
from pywaclient.endpoints.manuscripts import ManuscriptCrudEndpoint
from pywaclient.endpoints.map_marker_types import MapMarkerTypeCrudEndpoint
from pywaclient.endpoints.maps import MapCrudEndpoint
from pywaclient.endpoints.notebooks import NotebookCrudEndpoint
from pywaclient.endpoints.rpg_system import RpgSystemCrudEndpoint
from pywaclient.endpoints.secrets import SecretCrudEndpoint
from pywaclient.endpoints.subscriber_groups import SubscriberGroupCrudEndpoint
from pywaclient.endpoints.timelines import TimelineCrudEndpoint
from pywaclient.endpoints.users import UserCrudEndpoint
from pywaclient.endpoints.variables import VariableCollectionCrudEndpoint
from pywaclient.endpoints.worlds import WorldCrudEndpoint


class BoromirApiClient:

    def __init__(self,
                 name: str,
                 url: str,
                 version: str,
                 application_key: str,
                 authentication_token: str,
                 ):
        self.headers = {
            'x-auth-token': authentication_token,
            'x-application-key': application_key,
            'Accept': 'application/json',
            'User-Agent': f'{name} ({url}, {version})'
        }
        self.headers_post = self.headers.copy()
        self.headers_post['Content-type'] = 'application/json'
        self.base_url = 'https://www.worldanvil.com/api/external/boromir/'
        self.block = BlockCrudEndpoint(self)
        self.block_folder = BlockFolderCrudEndpoint(self)
        self.article = ArticleCrudEndpoint(self)
        self.image = ImageCrudEndpoint(self)
        self.manuscript = ManuscriptCrudEndpoint(self)
        self.user = UserCrudEndpoint(self)
        self.secret = SecretCrudEndpoint(self)
        self.world = WorldCrudEndpoint(self)
        self.category = CategoryCrudEndpoint(self)
        self.variable_collection = VariableCollectionCrudEndpoint(self)
        self.rpg_system = RpgSystemCrudEndpoint(self)
        self.subscriber_group = SubscriberGroupCrudEndpoint(self)
        self.map = MapCrudEndpoint(self)
        self.map_marker_types = MapMarkerTypeCrudEndpoint(self)
        self.history = HistoryCrudEndpoint(self)
        self.timeline = TimelineCrudEndpoint(self)
        self.canvas = CanvasCrudEndpoint(self)
        self.chronicle = ChronicleCrudEndpoint(self)
        self.block_template = BlockTemplateCrudEndpoint(self)
        self.notebook = NotebookCrudEndpoint(self)


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    test_article_1 = client.article.put(
        {
            'title': 'Secret Article',
            'templateType': 'article',
            'world': {
                'id': world_id}
        }
    )
    test_article_2 = client.article.put(
        {
            'title': 'Secret Article 2',
            'templateType': 'article',
            'world': {
                'id': world_id
            }
        }
    )

    test_subscriber_group_1 = client.subscriber_group.put(
        {
            'title': 'A Secret Subscriber Group',
            'world': {
                'id': world_id
            }
        }
    )

    test_subscriber_group_2 = client.subscriber_group.put(
        {
            'title': 'A Secret Subscriber Group 2',
            'world': {
                'id': world_id
            }
        }
    )

    test_subscriber_group_3 = client.subscriber_group.put(
        {
            'title': 'A Secret Subscriber Group 3',
            'world': {
                'id': world_id
            }
        }
    )

    test_secret_1 = client.secret.put(
        {
            'title': 'Test Secret Creation',
            "content": "This is some content for the secret.",
            "article": {
                "id": test_article_1['id']
            },
            'world': {
                'id': world_id
            },
            'subscribergroups': [
                {
                    'id': test_subscriber_group_1['id']
                }
            ]
        }
    )
    test_secret_2 = client.secret.put(
        {
            'title': 'Test Secret Creation 2',
            'world': {
                'id': world_id
            }
        }
    )
    response_patch_secret_2 = client.secret.patch(
        test_secret_2['id'],
        {
            'title': 'New Title',
            'content': 'The new content for an secret.',
            'article': {
                'id': test_article_1['id']
            },
            'subscribergroups': [
                {
                    'id': test_subscriber_group_2
                }
            ],
            'tags': 'some,other,tags'
        }
    )

    full_test_secret_2 = client.secret.get(
        test_secret_2['id'],
        1
    )
    assert full_test_secret_2['title'] == 'New Title'
    assert full_test_secret_2['content'] == 'The new content for an secret.'
    assert full_test_secret_2['tags'] == 'some,other,tags'
    assert full_test_secret_2['article']['id'] == test_article_1['id']
    assert full_test_secret_2['subscribergroups'][0]['id'] == test_subscriber_group_2['id']
    assert len(full_test_secret_2['subscribergroups']) == 1

    full_test_secret_2['subscribergroups'].append(
        {
            'id': test_subscriber_group_3['id']
        }
    )
    client.secret.patch(
        test_secret_2['id'],
        {
            'subscribergroups': full_test_secret_2['subscribergroups']
        }
    )
    full_test_secret_2 = client.secret.get(
        test_secret_2['id'],
        1
    )

    print(full_test_secret_2)
    assert len(full_test_secret_2['subscribergroups']) == 2

    client.secret.delete(test_secret_1['id'])
    client.secret.delete(test_secret_2['id'])

    client.article.delete(test_article_1['id'])
    client.article.delete(test_article_2['id'])

    client.subscriber_group.delete(test_subscriber_group_1['id'])
    client.subscriber_group.delete(test_subscriber_group_2['id'])
    client.subscriber_group.delete(test_subscriber_group_3['id'])


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from collections import Counter

from init_client import client
from pywaclient.exceptions import UnprocessableDataProvided

if __name__ == '__main__':
    wbtv = 'a86b6da9-6ba2-413a-87d9-ee98cbc6d9b9'

    new_world = client.world.put({
        'title': 'A New Horizon'
    })
    print(new_world)
    client.world.delete(new_world['id'])

    result = client.world.get(wbtv, 2)
    counter = Counter()
    try:
        for a in client.world.statblock_folders(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.blocks(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    try:
        count = 0
        for a in client.world.articles(wbtv, category_id='061b880e-0f69-4eae-8b64-954b56ae26b4'):
            count += 1
            print(a['title'])
        print("Category article: ", count)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    try:
        count = 0
        for a in client.world.articles(wbtv, category_id='-1'):
            count += 1
            print(a['title'])
        print("Top level articles:", count)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.articles(wbtv):
            counter.update([a['id']])
        print("All articles", len(counter))
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.subscriber_groups(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.categories(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.secrets(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.chronicles(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.timelines(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.histories(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.maps(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.manuscripts(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.images(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.canvases(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)

    counter = Counter()
    try:
        for a in client.world.variable_collections(wbtv):
            counter.update([a['id']])
        print(counter)
    except UnprocessableDataProvided as err:
        print(err.status)
        print(err.error_tracestack)



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable
from pywaclient.endpoints import CrudEndpoint


class BlockCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'block')


class BlockFolderCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'blockfolder')
        self.path_blocks = 'blockfolder/blocks'

    def blocks(self, block_folder_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        Retrieve a list of blocks which are part of a specific block.

        :param block_folder_id:    The id of the world the blocks should be returned from.
        :param complete:    Returns all the folders as an iterable.
        :param limit:       Determines how many articles are returned. Value between 1 and 50.
        :param offset:      Determines the offset at which articles are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            return self._scroll_collection(self.path_blocks, {'id': block_folder_id}, 'entities')
        return self._post_request(self.path_blocks,
                                  {'id': block_folder_id},
                                  {'limit': limit, 'offset': offset})['entities']





#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Iterable, Dict, Any

from pywaclient.endpoints.eras import EraCrudEndpoint
from pywaclient.endpoints import CrudEndpoint


class TimelineCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'timeline')
        self.path_eras = 'eras'
        self.era = EraCrudEndpoint(client)

    def eras(self, world_id: str, timeline_id: str, complete: bool = True, limit: int = 50,
                  offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        List all eras by a timeline given, filtered with a limit of entities shown and an offset.

        :param world_id:                The id of the world the eras should be returned from.
        :param timeline_id:             The id of the timeline to return the eras from.
        :param complete                 Ignore limit and offset and return all the eras as an iterable. Will fetch a new batch
                                        every 50 variables.
        :param limit:                   Determines how many eras are returned. Value between 1 and 50.
        :param offset:                  Determines the offset at which eras are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            return self._scroll_collection(self.path_eras, {'id': world_id}, 'entities', 'timeline',
                                           timeline_id)
        return self._post_request(self.path_eras,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset,
                                   'timeline': {'id': timeline_id}})['entities']



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    test_variableCollection_1 = client.variable_collection.put(
        {
            "title": "Variables Collection",
            "description": "All variables used in this world.",
            "world": world_id,
            "prefix": "test-prefix",
            "state": "public"
        }
    )
    test_variable_1 = client.variable.put(
        {
            "k": "key",
            "v": "value",
            "type": "string",
            "collection": test_variableCollection_1['id'],
            "world": world_id
        }
    )
    response_patch_variable_1 = client.variable.patch(
        test_variable_1['id'],
        {
            'v': 'This is a new value for the variable.'
        }
    )

    full_test_variable_1 = client.variable.get(
        test_variable_1['id'],
        2
    )

    assert full_test_variable_1['v'] == 'This is a new value for the variable.'

    client.variable_collection.delete(test_variableCollection_1['id'])
    client.variable.delete(test_variable_1['id'])



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    test_chronicles_1 = client.chronicle.put(
        {
            "title": "Chronicle of the World",
            "world": {
                "id": world_id
            }
        }
    )
    test_chronicles_2 = client.chronicle.put(
        {
            'title': 'Test Chronicles Creation 2',
            'templateType': 'chronicles',
            'world': {
                'id': world_id}
        }
    )
    response_patch_chronicles_2 = client.chronicle.patch(
        test_chronicles_2['id'],
        {
            'excerpt': 'This is an excerpt for an chronicles.'
        }
    )

    full_test_chronicles_2 = client.chronicle.get(
        test_chronicles_2['id'],
        2
    )

    assert full_test_chronicles_2['excerpt'] == 'This is an excerpt for an chronicles.'

    client.chronicle.delete(test_chronicles_1['id'])
    client.chronicle.delete(test_chronicles_2['id'])

    chronicles_with_a_lot_of_views = client.chronicle.put(
        {
            'title': 'An chronicles with a lot of views.',
            'templateType': 'chronicles',
            'world': {
                'id': world_id
            }
        }
    )
    print(client.chronicle.get(chronicles_with_a_lot_of_views['id'], 2))



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Iterable

from pywaclient.endpoints import CrudEndpoint, download_binary


class ImageCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'image')

    def get_binary(self, identifier: str) -> Iterable[bytes]:
        """Get the image binary by identifier

        :param identifier: Identifier of the image.
        :return: Iterable of image binary chunks.
        """
        metadata = self.get(identifier, -1)
        return download_binary(metadata['url'])



import setuptools

with open('README.md', 'r') as fp:
    setuptools.setup(
        name="pywaclient",
        use_scm_version=True,
        setup_requires=['setuptools_scm'],
        author="Jonas Waeber",
        author_email="jonaswaeber@gmail.com",
        description="A simple wrapper client library for the World Anvil API.",
        long_description=fp.read(),
        long_description_content_type="text/markdown",
        url="https://gitlab.com/SoulLink/world-anvil-api-client",
        packages=['pywaclient', 'pywaclient.endpoints', 'pywaclient.exceptions'],
        install_requires=['requests'],
        license='Apache 2.0',
        classifiers=[
            "Programming Language :: Python :: 3.7"
        ],
    )



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':

    manuscript_1 = client.manuscript.put(
        {
            'title': 'Test Manuscript Creation',
            'world': {
                'id': world_id}

        })

    client.manuscript.tag.put(
        {
            'manuscript': {
                'id': manuscript_1['id']
            },
            'title': 'Test Manuscript Tag Creation',
        })

    for manuscript_tag in client.manuscript.tags(manuscript_1['id']):
        print(manuscript_tag)

    client.manuscript.label.put(
        {
            'manuscript': {
                'id': manuscript_1['id']
            },
            'title': 'Test Manuscript Label Creation',
        })

    for manuscript_label in client.manuscript.labels(manuscript_1['id']):
        print(manuscript_label)

    # for manuscript_bookmark in client.manuscript.bookmarks(manuscript_1['id']):
    #     print(manuscript_bookmark)

    client.manuscript.version.put(
        {
            'manuscript': {
                'id': manuscript_1['id']
            },
            'title': 'Test Manuscript Version Creation',
        }
    )



    for manuscript_version in client.manuscript.versions(manuscript_1['id']):
        print(manuscript_version)

        client.manuscript.version.part.put(
            {
                'version': {
                    'id': manuscript_version['id']
                },
                'type': 'folder',
                'title': 'Test Manuscript Part Folder Creation',
            }
        )
        client.manuscript.version.part.put(
            {
                'version': {
                    'id': manuscript_version['id']
                },
                'type': 'richMediaUrl',
                'title': 'Test Manuscript Part Image Creation',
            }
        )
        client.manuscript.version.part.put(
            {
                'version': {
                    'id': manuscript_version['id']
                },
                'type': 'text',
                'title': 'Test Manuscript Part Image Creation',
            }
        )

        for manuscript_part in client.manuscript.version.parts(manuscript_version['id']):
            print(client.manuscript.version.part.get(manuscript_part['id'], 2))

            client.manuscript.version.part.beat.put(
                {
                    'part': {
                        'id': manuscript_part['id']
                    },
                    'title': 'Test Manuscript Beat Creation',
                }
            )

            for manuscript_beat in client.manuscript.version.part.beats(manuscript_part['id']):
                print(client.manuscript.version.part.beat.get(manuscript_beat['id'], 2))



        for manuscript_plot in client.manuscript.version.plots(manuscript_version['id']):
            print(manuscript_plot)

        for manuscript_stat in client.manuscript.version.stats(manuscript_version['id']):
            print(manuscript_stat)
    print(manuscript_1)
    full_manuscript_1 = client.manuscript.get(manuscript_1['id'], 3)

    print(full_manuscript_1)
    client.manuscript.delete(manuscript_1['id'])



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    test_article_1 = client.article.put(
        {
            'title': 'Secret Article',
            'templateType': 'article',
            'world': {
                'id': world_id}
        }
    )
    test_history_1 = client.history.put(
        {
            'title': 'Test History Creation',
            'year': 1000,
            'world': {
                'id': world_id
            }
        }
    )

    # test_category_1 = client.category.put(
    #     {
    #         'title': 'Test Category Creation',
    #         'world': {
    #             'id': world_id
    #         }
    #     }
    # )

    test_subscriber_group_1 = client.subscriber_group.put(
        {
            'title': 'A Secret Subscriber Group',
            'world': {
                'id': world_id
            }
        })



    test_timelines_1 = client.timeline.put(
        {
            'title': 'Test Timelines Creation',
            'state': 'private',
            'world': {
                'id': world_id
            },
            'icon': 'fa-solid fa-calendar',
            'subscribergroups': [
                {
                    'id': test_subscriber_group_1['id']
                }
            ],
            'article': {
                'id': test_article_1['id']
            },
            # 'category': {
            #     'id': test_category_1['id']
            # },
            'tags': 'test,test2',
            'histories': [
                {
                    'id': test_history_1['id']
                }
            ],
            'description': 'Test Description',
            'utdOffset': 0,
            'type': 'parallel',
            'showInToC': True,
            'calendar': {
                'id': '102144'
            }
        }
    )

    test_era = client.timeline.era.put(
        {
            'title': 'Test Era Creation #1',
            'world': {
                'id': world_id
            },
            'timeline': {
                'id': test_timelines_1['id']
            }
        }
    )
    print(test_era)
    test_timelines_2 = client.timeline.put(
        {
            'title': 'Test Timelines Creation 2',
            'world': {
                'id': world_id
            }
        }
    )
    response_patch_timelines_2 = client.timeline.patch(
        test_timelines_2['id'],
        {
            'title': 'Change the title'
        }
    )

    full_test_timelines_2 = client.timeline.get(
        test_timelines_2['id'],
        2
    )

    assert full_test_timelines_2['title'] == 'Change the title'

    client.timeline.delete(test_timelines_1['id'])
    client.timeline.delete(test_timelines_2['id'])

    client.article.delete(test_article_1['id'])
    client.history.delete(test_history_1['id'])
    # client.category.delete(test_category_1['id'])
    client.subscriber_group.delete(test_subscriber_group_1['id'])




#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable

from pywaclient.endpoints import CrudEndpoint


class MapMarkerCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'marker')


class MapMarkerGroupCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'markergroup')
        self.marker = MapMarkerCrudEndpoint(client)
        self.path_markers = 'markergroup/markers'

    def markers(self, map_maker_group_id: str) -> Iterable[Dict[str, Any]]:
        """Iterate over all map markers in a map marker group.
        """
        return self._scroll_collection(self.path_markers, {'id': map_maker_group_id}, 'entities')


class MapLayerCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'layer')


class MapCrudEndpoint(CrudEndpoint):

    def __init__(self, client: 'BoromirApiClient'):
        super().__init__(client, 'map')
        self.path_layers = f'{self.path}/layers'
        self.layer = MapLayerCrudEndpoint(client)
        self.path_groups = f'{self.path}/markergroups'
        self.marker_group = MapMarkerGroupCrudEndpoint(client)
        self.path_markers = f'{self.path}/markers'
        self.marker = MapMarkerCrudEndpoint(client)

    def marker_groups(self, map_id: str) -> Iterable[Dict[str, Any]]:
        """
        List the marker groups of a specific map.

        :param map_id:      The id of the map.
        :return:            Iterable of all marker groups.
        """
        return self._scroll_collection(self.path_groups, {'id': map_id}, 'entities')

    def layers(self, map_id: str) -> Iterable[Dict[str, Any]]:
        """
        List the layers of a specific map.

        :param map_id:      The id of the map.
        :return:            Iterable of all layers.
        """
        return self._scroll_collection(self.path_layers, {'id': map_id}, 'entities')

    def markers(self, map_id: str) -> Iterable[Dict[str, Any]]:
        """
        List the markers of a specific map.

        :param map_id:      The id of the map.
        :return:            Iterable of all markers.
        """
        return self._scroll_collection(self.path_markers, {'id': map_id}, 'entities')
